	PAGE	55,132
	TITLE	WD8STAR2.ASM - WD8003 / StarLAN Hardware dependent routines

COMMENT \
/************************************************************************
* WD8STAR2.ASM								*
*									*
* (c) Copyright 1987 Western Digital Corporation (R).			*
* All right reserved. Contains confidential information and trade 	*
* secrets proprietary to Western Digital.				*
*									*
* This asm file and the MACMOD.C file will handle the Interface between *
* NETBIOS OSI modules and Western Digital WD8003 StarLAN PC Adapter	*
* Board.  								*
*									*
*									*
* FUNCTION:								*
* --------								*
*   There are five routines in this asm file				*
*									*
*   _init_8390	initiate 8390 NIC. initiate the LAN controller registers*
*									*
*   _hdw_isr	Services all LAN hardware interrupts.			*
*									*
*   _cp_frm 	copy a frame in system memory to tx buff of shared 	*
*		memory. (argument: ptr to the frame buffheader)		*			*
*   _xmt_frm 	initiate a tx of frame in shared memory			*
*   _rcv_frm	copy a rx frame from rx buff ring of shared memory to	* 				
* 		system memory and put it into _rx_queue			*
* 		argument: ptr to the frm in ring 			*
*									*
*									*
*    1. Base i/o address of board may be specified on driver command	*
*	line. The Address ROM checksum is used to verify that the	*
*	specified base i/o address is accurate. If not entered on	*
*	command line, default will be 280h.				*
*    2. A 6 byte LAN address can be specified on the driver command	*
*	line. If entered, it will override the address contained in the *
*	LAN Address ROM on the WD8000. The value entered with the	*
*	switch must be 6 hexadecimal bytes in the following order: 	*
*	xx xx xx yy yy yy; where the 3 bytes described as "xx" are the	*
*	board address, and the 3 bytes described as "yy" are Globally	*
*	Assigned Address Block. Note that the leftmost (first) byte is	*
*	the least significant byte of the LAN address and the rightmost *
*	(last) byte must be the most significant.			*
*    3. The global '_phreaque' is used to flag a catastrophic error,	*
*									*
*									*
* HISTORY:								*
* -------								*
*									*
*	DATE:   3/18/87							*
* 	LEVEL:  et.bb.00 						*
*	WHO:    KC							*
*	CHANGE:	create file for wd8003					*
*									*
*	DATE:   6/18/87							*
* 	LEVEL:  et.bb.01 						*
*	WHO:    KC	    	 					*
*	CHANGE:	WDBPR# 762						*
*	[01]	NETBIOS does not block hardware interrupt during 	*
*		intialization. 						*
*									*
*	DATE:   7/18/87							*
* 	LEVEL:  et.bb.02 						*
*	WHO:    KC							*
*	PROBLEM:8003E will rec frame with frame with huge data count	*
*		the CRC check did not detect it (?, must be a hw bug)	*
*	CHANGE:	[02] WDBPR# 777          				*
*		in RCV_FRM check the byte count before copy frame from  *
*		shared mem into OSI buffer. If frame length is too      *
*		large (>_bufsize 0x22c=566), drop the frame.		*
*		a. increment the _phreaque cnt if above error occurs	*
*		b. also remove the patch error				*
*									*
*	DATE:   9/24/87							*
* 	LEVEL:  et.bb.05 						*
*	WHO:    KC							*
*	PROBLEM:isr need to follow a specific sequence to recover from	*
*		buffer ring overflow condition.                      	*
*	CHANGE:	[03] WDBPR# 818          				*
*		in _hdw_isr(), if ovw set in interrupt status register   * 
*		go through the recovery sequence according the NATIONAL *
*									*
*	DATE:   9/24/87							*
* 	LEVEL:  et.bb.05 						*
*	WHO:    KC							*
*	PROBLEM:8390 has rcv byte swap porblem, the byte counte check 	*
*		at _rcv_frm has to be correct in order to prevent data	*
*		segment from being corrupted.				*
*	CHANGE:	[04] WDBPR# 822          				*
*	       	in _rcv_frm() check frame size against max buff size,	*
*		drop frame if too large. Use JBE instead of JLE.	*
*									*
*	DATE:   9/24/87							*
* 	LEVEL:  et.bb.05 						*
*	WHO:    KC							*
*	PROBLEM:add delay before tx frame in order to run 8003S with  	*
*		8000S.							*
*	CHANGE:	[05] WDBPR# 823          				*
*	       	a. add routine starlan_wait() to set up wait loop count *
*		b. _xmt_frm() do loop delay before tx frame		*
*
*	WHO:  Steve Chen						*
*	DATE: 10/7/87
*	CHANGE: Porting to Microsoft C compiler 4.0
*
*	WHO: Steve Chen [6]
*	DATE: 3/16/88
*	CHANGE: Make changes, so it can run on micro channel machine	*
*		also.    
*************************************************************************\

	SUBTTL	Include Files
	PAGE
;*******************************************************************************
; INCLUDE FILES
;*******************************************************************************
;
;
DGROUP	GROUP	_DATA
_DATA	SEGMENT WORD PUBLIC 'DATA'
;
	INCLUDE WD8STAR2.equ
	INCLUDE frameblk.equ

	PUBLIC  _bfrovw			; [03]
	PUBLIC	_phreaque
	PUBLIC	_my_eaddr
	PUBLIC	_WD8_base
	PUBLIC	_SM_base
	PUBLIC	_bufsize
	PUBLIC	_adapter_irq
	PUBLIC  _in_diags	; [01]
; /* %%5%%+ */
	PUBLIC	_m_channel	; [06] micro channel flag
; /* %%5%%- */
; /* %%4%%+ */
	PUBLIC	_int_reg	; [06]
; /* %%4%%- */
;
; defined variable and reserve patch area 

;***********************************************************************
;
; DRIVER PATCH AREA
;***********************************************************************

;drv_patch	db	1024 DUP (0)          ;[02]

;***********************************************************************
;
;	Miscellaneous Variables
;***********************************************************************

; buffer ring overflow counter
_bfrovw	  dw 	0 		; [03]
;
_phreaque  db	00h		; Catastrophic error on LAN
;
_my_eaddr  dw	0			; Pointer to 6 byte LAN address
;
; Command line LAN address parameter is entered as xxxxxxyyyyyyh
; where xxxxxx is 3 byte board address and yyyyyy is 3 byte GAAB (vendor id)
; and is stored as:
;                   +-----------------------------------------+
;                   | lsb  | ---- | msb  | lsb  | ---  | msb  |
;                   +-----------------------------------------+
;		      ^			    ^
;  _my_eaddr	      |			    |
;	  board addr:-+			    |
;  					    |
;         GAAB will			    |
;	  start here:-----------------------+
;
;
_WD8_base dw	0280h			; Base i/o address of board
_SM_base  dw	0C400h			; BASE addr of shared memory default
					; to C400h (paragraph number)	
; /* %%5%%+ */
_m_channel 	dw 	0		;[6] micro channel flag	
; /* %%5%%- */
; /* %%4%%+ */
_int_reg	dw	0		;[6] interrupt register
; /* %%4%%- */

;
; Used by init_lanc() as flags and temporary storage
;
_bufsize	dw	0			; Size of buffers in this implementation
;
; Used to track any spurious interrupts generated.
;
spur_evt db	0				; Event causing int
; /* %%3%%+ */
stop_page	db	32			; [6]stop page # 
; /* %%3%%- */
;
; Used during normal operation
;
rc	dw	0			; Return Code
stalen	dw	0			; DEBUG
int_ents  dw	0			; Counter incremented on entry to ISR
cur_txbuf dw	0			; Address of current TX buffer

_in_diags db	0			; != 0 if in diagnostic routines
					;		 (hdw_init)
_adapter_irq	db	0		; IRQ to be used with WD board [00.n]

; The client's context is stored on our stack for device
; interrupts since these are asynchronous and we don't have to
; return or get any information.
;
; *  Note: these will be set up in OSI/NETBIOS rather than here....
; _bstack	db 80H DUP(?)	; the stack
; _bstacktop	dw	0	; top of the stack to stuff into sp

_DATA	ENDS
	PAGE
;
_TEXT	SEGMENT BYTE PUBLIC 'CODE'
	ASSUME	CS:_TEXT, DS:DGROUP


	EXTRN	_enter_osi:NEAR		; routine to do background
	EXTRN	_BuffAlloc:NEAR		; Allocates buffer
	EXTRN	_BuffFree:NEAR		; Allocates buffer
	EXTRN	_QInsert:NEAR		; routine to insert buffer onto queue
	EXTRN	_QRemove:NEAR		; routine to take TX buff off queue
	EXTRN	_datsav:WORD		; DS segment register - used by ISR


	SUBTTL	Macros
	page
;***********************************************************************
;
;	Macros
;
;***********************************************************************

;
; MACRO rd_wd8
;   Reads port specified on macro call. Leaves byte in AL.
;

rd_wd8	MACRO	port
	mov	DX, _WD8_base
	add	DX, port		; DX contains address of port
	in	AL, DX			; AL contains data read from port
	ENDM

;
; MACRO wr_wd8
;   Writes byte in AL to port specified on macro call.
;

wr_wd8	MACRO	port
	mov	DX, _WD8_base
	add	DX, port		; DX contains address of port
	out	DX, AL			; AL contains data to be written to port
	ENDM


; [05] add new _routine starlan_wait
	subttl	_starlan_wait - set LAN type and Starlan write wait states
	page
; ********************************************************************
;_STARLAN_WAIT							     *
;								     *
;								     *
; Function:							     *
;								     *
; This routine will initialize LAN type variable "starlan"; it will  *
; also determine the type of host, and if non-PC or non-XT, will set *
; a starlan write wait state variable to 500 micro-second delay      *
; based on a 16MHz Compaq.					     *
;								     *
; C Call:	starlan_wait by macmod				     *
;								     *
; Calls:							     *
;								     *
; On Entry:							     *
;								     *
; On Exit:							     *
;		AX, BX, DX	corrupted			     *
;								     *
; Procedure:							     *
;								     *
; NOTES:							     *
;								     *
; ********************************************************************

ADDROM	equ	8			;offset from base address to LAN address ROM
IDBYTE	equ	6			;ID byte at offset 6 in LAN address ROM
starlan db	0			;LAN type: 0=ethernet , !0=starlan

	public	_starlan_wait

_starlan_wait proc near

	mov	dx,_WD8_base		;if ((starlan = in(ADDROM + _WD8_base + 6)) == 2) {
	add	dx,ADDROM+IDBYTE
	in	al,dx
	mov	cs:starlan,al
	cmp	al,0
	je	starlan_wait99

;	ROM ID byte is at F000:FFFE = FF = PC
;				    = FE = XT
;				    = FD = Jr
;				    = FC = AT class machines (including COMPAQ 386)

	push	es
	mov	ax,0f000h
	mov	es,ax
	mov	bx,0fffeh
	mov	al,es:[bx]
	and	al,0feh 		;if (!PC or XT)
	cmp	al,0feh
	je	starlan_wait10
	mov	cs:starlan_waitloopcnt,726 ;(500 micro-seconds on a 16MHz 386)
	jmp	short starlan_wait10

starlan_waitloopcnt  dw  1			 ;at least 1 loop

starlan_wait10:  pop	 es

starlan_wait99:  ret
_starlan_wait	 endp
; end [05]



	SUBTTL _init_8390
	PAGE
; ********************************************************************
; INIT_8390()							     *
;								     *
;								     *
; Function:							     *
;								     *
; This routine will initialize 8390. set up the rcv buff ring and tx *
; buf flag (_tb_cur, _tb_use). 					     *
;								     *
; C Call:	_init_8390() by init_lanc, stop_mac		     *
;								     *
; Calls:							     *
;								     *
; On Entry:							     *
;								     *
; _WD8_base, _bufsize, _my_eaddr already set up by init_lanc	     *
;								     *
; On Exit:							     *
; 		AX, BX, CX, DX	corrupted			     *
; 								     *
;								     *
; Procedure:							     *
;								     *
; 								     *
; NOTES:							     *
;								     *
;	1. set FIFO threshold to 8 bytes by set BMS=1 and FT1=1 FT0=0*
;	   in Data Configuration Register.			     *
;								     *
; ********************************************************************

	PUBLIC  _init_8390		; initialize 8390

_init_8390 PROC	NEAR


	cld				; Clear direction flag for movs...

; /* %%3%%+ */
					;[6] 16K share memory for micro channel
	mov	stop_page, STOP_PG 
	cmp	_m_channel,0	
	je	init1 			;not micro channel machine
	mov	stop_page, STOP_PG + STOP_PG	;micro channel machine
init1:	
; /* %%3%%- */
;
; initial the LAN Controller register
;					

; program for page 0
	mov	AL, MSK_PG0 + MSK_RD2
	wr_wd8	CMDR
; initial DCR data configuration
					;
; /* %%2%%+ */
	mov	AL, MSK_BMS + MSK_FT10	; select FIFO threshold = 8 bytes
	cmp	_m_channel,0		; [6]
	je	init2			; not micro channel machine
	mov	AL, MSK_BMS + MSK_FT10 + MSK_WTS  	;word mode
init2:
	wr_wd8	DCR
; /* %%2%%- */
; clr RBCR0,1
	xor	AL, AL
	wr_wd8	RBCR0
	wr_wd8	RBCR1
; initial RCR to monitor mode
	mov	AL, MSK_MON
	wr_wd8	RCR			; disable the rxer
; initial TCR
	xor	AL, AL			
	wr_wd8	TCR			; normal operation
; intial rev buffer ring
; /* %%3%%+ */
	mov	AL, stop_page		;[6] 
	wr_wd8	PSTOP			; init PSTOP
; /* %%3%%- */
	mov	AL, STRT_PG		
	wr_wd8	PSTART			; init PSTART to the 1st page of ring
	wr_wd8	BNRY			; init BNRY
; clr ISR by 1's
	mov	AL, -1 			; write FF
	wr_wd8	ISR
; initial IMR
	mov	AL, 3Fh
	wr_wd8	IMR			; enable interrupt
; program for page 1
	mov	AL, MSK_PG1 + MSK_RD2
	wr_wd8	CMDR
; inital physical addr
	mov	CX, BPNA
	mov	BX, _my_eaddr		; ptr to adr in BX
	mov	DX, _WD8_base
	add	DX, PAR0		; i/o address of PAR0 in DX
lopa:
	mov	AL, [BX]		; get 1 byte into AL
	out	DX, AL			; write to PAR
	inc	BX
	inc	DX
	loop	lopa
; inital multicast filter,  write all 0's  into MAR0 - MAR7
	mov	CX, 8
	mov	DX, _WD8_base
	add	DX, MAR0		; i/o address of MAR0 in DX
	xor	AL, AL			
lopb:
	out	DX, AL
	inc	DX
	loop	lopb					
; initial CURR = PSTART + 1
	mov	AL, STRT_PG + 1
	wr_wd8	CURR
; program for page 0
	mov	AL, MSK_PG0 + MSK_RD2
	wr_wd8	CMDR
; if call by stop_mac, do not put it on line
	cmp	_OSI_RST, 1
	je	skip_it			; call by stop_mac
	
; put 8390 on line
	mov	AL, MSK_STA + MSK_RD2		; activate 8390
	wr_wd8	CMDR
; program RCR to normal operation (MSK_AB, no MSK_AM)
	mov	AL, MSK_AB			; accept broadcast
	wr_wd8	RCR
	
;
; might as well reset the tx buf of the shr mem
;
skip_it:
	XOR	AL, AL
	mov	_tb_use, AL
	mov	_tb_cur, AL
		
; return now
	ret

_init_8390 ENDP







        SUBTTL _hdw_isr
        PAGE
; ***************************************************************
; _hdw_isr()
;
; Function:
;
; Service interrupt from the 8390 via the 8003 board. The 
; interrupt will be serviced in the order of
;    1. overwrite warning (OVW).
;    2. RX good (PRX), and RX with error (RXE).
;    3. TX good (PTX).
;    4. TX with error (TXE).
;    5. counter overflow (CNT).
;
; Procedure:
;
;    save registers
;    set up new stack
;    set up addressability to data.
;    send EOI to 8259.
;    
;    WHILE (read ISR & 0x3f != 0)
;
;         IF (OVW=1)
;              report error
;              jump to receive process
;
;         ELSEIF (PRX=1 or RXE=1)
;              WHILE (RCV BUFF RING not empty)
;                   read byte 0 of next frame pointed by BNRY
;                   IF (PRX set: status good)
;                        call _rcv_frm()
;                   ELSE discrad frame by forwarding BNRY
;                   ENDIF
;              ENDWHILE
;              clear PRX, RXE, OVW of ISR by 1's
;
;         ELSEIF (PTX=1 or TXE=1)
;              IF (PTX=1)
;                   Incr .txok
;                   update .collsn with NCR if COL=1
;                   Incr .lostcrs if CRS=1
;                   clear PTX of ISR
;              ELSE                           /* TXE=0 */
;                   Incr .txbad
;                   Incr .underrun of FU=1
;                   clear TXE of ISR
;              ENDIF
;              _tb_use --
;              _tb_cur = (_tb_cur+1)%n   /* n is no. of tx buf in shard */
;              IF (_tb_use>0)
;                   _xmt_frm()
;                   WHILE(tx_q not empty & _tb_use <n)
;                        Qremove(_tx_buf)
;                        _cp_frm(_tx_buf)
;                        _tb_use ++
;                   ENDWHILE
;              ELSEIF (tx_q not empty)
;                   Qremove(_tx_buf)
;                   _cp_frm(_tx_buf)
;                   _tb_use ++
;                   xmr_frm()
;                   WHILE(tx_q not empty & _tb_use <n)
;                        Qremove(_tx_buf)
;                        _cp_frm(_tx_buf)
;                        _tb_use ++
;                   ENDWHILE
;              ENDIF
;
;         ELSEIF (CNT=0)
;              update .algerr by CNTR0
;              update .crcerr by CNTR1
;              update .???    by CNTR2 for missed frm
;              clear CNT of ISR
;
;         ENDIF
;
;    ENDWHILE
;
;    restore stack
;    call _enter_osi()
;    restore registers
;
;								 
; Calls:
;	 
; _cp_frm
; _rcv_frm
; _xmt_frm
; _enter_osi
; Qremove
;
;
; ***************************************************************


	PUBLIC	_hdw_isr			; 8003 board irq service routine
	PUBLIC 	_phreaque

iintrss	dw	0			; save ss
iintrsp dw	0			; save sp

_hdw_isr	PROC	FAR

; save registers
	push 	AX
	push	BX
	push	CX
	push	DX
	push	DS
	push	ES
	push	BP
	push    DI
	push	SI

; set up new stack
	mov	CS:iintrss, ss		; save their stack
	mov	CS:iintrsp, sp
	mov	SS, CS:_datsav		; set up our stack
	mov	SP, OFFSET DGROUP:_bstacktop

	cld				; clr direction flag

; set up addressability, set up DS & ES
	mov 	AX, CS:_datsav
	mov	DS, AX
	mov	ES, AX

	push	rc			; save rc for FG ???
; /* %%4%%+ */
; clear PIC 8259			; [6]send EOI just before exit
;	mov	AL, EOI			; this is for level sensitive int 
;	mov	DX, INTA00
;	out	DX, AL
; /* %%4%%- */

; read irq status register
rd_isr:
	rd_wd8	ISR			; read isr into AL
	and	AL, 3Fh			; check bit0-bit5, if all 1's no irq
	cmp	AL, 0			; if any irq
	jne	tst_ovw			; some irq
	jmp	isr_exit		; no more irq, exit

; process OVW	(OVW = 1)       
;    may report error here
tst_ovw:
	test	AL, MSK_OVW		; if OVW irq
	jnz	prcs_ov			; OVW (OVW = 1) [03]
	jmp	test_rx			; no OVW (OVW = 0) [03]
;
; **************************************************************
; [03] start
; follow the DP8390 datasheet addendum to handle the buff ring overflow
;	 
prcs_ov:
; 1. issue a STOP mode command
	mov	AL, MSK_STP + MSK_RD2
	wr_wd8	CMDR
; 6. remove one packet from the ring
	rd_wd8	BNRY			; BNRY in AL
	add	AL, 1			; start page of frm in AL
; /* %%3%%+ */
	cmp	AL, stop_page		; [6]check boundary
; /* %%3%%- */
	jne	get1
	mov	AL, STRT_PG		
; ring not empty
get1:
	mov	BH, AL
	xor	BL, BL			; BX has the rx_frm pointer
	push	BX			; save the frm ptr
        mov	AX, _SM_base		
	mov 	ES, AX			; EX has the shr seg paragraph
	mov	AX, ES:[BX]		; [6]AL has the status byte
	test	AL, SMK_PRX		; if rx good
	jz	fd_bnr			; rx error, drop frm by forward bnry
; good frm, call _rcv_frm
	mov	BP, SP			; rx good
	push	BX			; pnter to rx frm buf ring as prm1
	call 	_rcv_frm   		; call _rcv_frm(pointer)			
	mov	SP, BP
	
; drop bad frame by forwarding the BNRY register
fd_bnr:					; drop frm by forward BNRY
; /* %%1%%+ */
	pop	BX			; restore frm ptr in BX
;	add	BX, 1			;[6]
	mov	ax, es:[bx]
	xchg	ah, al			;al = next pointer 
;	mov	AL, ES:[BX]		; next frm start page in AL
	sub	AL, 1			; new BNRY in AL
	cmp	AL, STRT_PG		; check boundary
; /* %%1%%- */
; /* %%3%%+ */
	jge	wrbnr
	mov	AL, stop_page		;[6]
	dec	AL			;[6] 
; /* %%3%%- */
wrbnr:
	wr_wd8	BNRY
; 2. clear the remote byte count registers (RBCR0,RBCR1)
	xor	AL, AL
	wr_wd8	RBCR0
	wr_wd8	RBCR1
; 3. poll the ISR for the RST bit
plisr:
	rd_wd8	ISR
	test	AL, MSK_RST
	jz	plisr		; keep polling until the RST bit set
; 4. place the NIC in loopback mode (mode 1 or 2) by writing 02 or 04 to TCR
	mov	AL, 02h		; put it in mode 2 (internal loopback)
	wr_wd8	TCR
; 5. issue start mode command
	mov	AL, MSK_STA + MSK_RD2
	wr_wd8	CMDR
; 7. out from loopback mode by writing 00 to TCR
	xor	AL, AL
	wr_wd8	TCR		; normal operation configuration
; clear OVW in ISR		      
	mov	AL, MSK_OVW 
	wr_wd8	ISR			; clear OVW
	inc	_bfrovw			; increment overflow counter
	jmp	rd_isr			; back to the top
;	
;
; end of the modification [03]
; *****************************************************

;	
;process PRX and RXE
;
test_rx:	    
	test  	AL, MSK_PRX	
	jnz	prcs_rx	     		; PRX = 1
	test 	AL, MSK_RXE		
	jnz	prcs_rx			; RXE = 1
	jmp	test_tx

prcs_rx:
	     
	mov	AL, MSK_PG1 + MSK_RD2	; read CURR reg
	wr_wd8	CMDR
	rd_wd8	CURR
	mov	BL, AL			; CURR in BL 
	mov	AL, MSK_PG0 + MSK_RD2		; read BNRY reg
	wr_wd8  CMDR
	rd_wd8	BNRY			; BNRY in AL
	add	AL, 1			; start page of frm in AL
; /* %%3%%+ */
	cmp	AL, stop_page		; [6]check boundary
; /* %%3%%- */
	jne	go_cmp
	mov	AL, STRT_PG		
go_cmp:
	cmp	AL, BL			
	je	end_rx			; buff ring empty
; ring not empty
	mov	BH, AL
	xor	BL, BL			; BX has the rx_frm pointer
	push	BX			; save the frm ptr
        mov	AX, _SM_base		
	mov 	ES, AX			; EX has the shr seg paragraph
; /* %%1%%+ */
	mov	AX, ES:[BX]		; [6]AL has the status byte
	test	AL, SMK_PRX		; if rx good
; /* %%1%%- */
	jz	fd_bnry			; rx error, drop frm by forward bnry
; good frm, call _rcv_frm
	mov	BP, SP			; rx good
	push	BX			; pnter to rx frm buf ring as prm1
	call 	_rcv_frm   		; call _rcv_frm(pointer)			
	mov	SP, BP
	
; drop bad frame by forwarding the BNRY register
fd_bnry:				; drop frm by forward BNRY
; /* %%1%%+ */
	pop	BX			; restore frm ptr in BX
;	add	BX, 1
	mov	ax, es:[bx]		; [6]
	xchg	ah, al			; al = next page pointer
;	mov	AL, ES:[BX]		; next frm start page in AL
	sub	AL, 1			; new BNRY in AL
	cmp	AL, STRT_PG		; check boundary
; /* %%1%%- */
	jge	wrbnry
; /* %%3%%+ */
	mov	AL, stop_page  		;[6]
	dec	AL			;[6]
; /* %%3%%- */
wrbnry:
	wr_wd8	BNRY
	jmp	prcs_rx	 
       
; clear PRX, RXE, OVW in ISR		      
end_rx:
	mov	AL, MSK_OVW + MSK_RXE + MSK_PRX
	wr_wd8	ISR			; clear OVW, RXE, PRX
	jmp	rd_isr			; back to the top

;	   
;process PTX and TXE
;
test_tx:
	test  	AL, MSK_PTX	
	jnz	prc_ptx	     		; PTX = 1
	test 	AL, MSK_TXE		
	jnz	prc_txe			; TXE = 1
	jmp	test_cnt
      
;
; process tx good, update txok, lostcrs, collsn
;			
prc_ptx:				; tx good 
	mov	SI, OFFSET DGROUP:_MACstat	; SI is the base
	add	WORD PTR [SI].txok, 1
	adc	WORD PTR [SI].txok + 2, 0	; update .txok 
	rd_wd8	TSR
	test	AL, SMK_CRS		; it crs set in TSR
	jz	nocrs			; no		       
	add	WORD PTR [SI].lostcrs, 1	; yes
	adc	WORD PTR [SI].lostcrs + 2, 0	; update .lostcrs
nocrs:	
	rd_wd8	NCR			; read number of collision in AL
	xor	AH, AH
	add	WORD PTR [SI].collsn, AX
	adc	WORD PTR [SI].collsn + 2, 0	; update .collsn
	mov	AL, MSK_PTX
	wr_wd8	ISR			; clear PTX
	jmp	tx_next

; 
; process tx error, update .txbad .underrun
;
prc_txe:				; tx bad
	mov	SI, OFFSET DGROUP:_MACstat	; SI is the base
	add	WORD PTR [SI].txbad, 1
	adc	WORD PTR [SI].txbad + 2, 0	; update .txbad 
	rd_wd8	TSR
	test	AL, SMK_FU		; it fu set in TSR
	jz	nofu			; no		       
	add	WORD PTR [SI].underrun, 1	; yes
	adc	WORD PTR [SI].underrun + 2, 0	; update .underrun
nofu:								  
    	mov	AL, MSK_TXE
	wr_wd8	ISR			; clear PTX
	jmp	tx_next

;
; tx next frame and fill up the tx buff in shared memory
;
tx_next:
	sub	_tb_use, 1		; buf in use decrement
	add	_tb_cur, 1		; forward the _tb_cur
	cmp	_tb_cur,tb_size		; check boundary condition
	jne	txnxt						  
	mov	_tb_cur, 0		
txnxt:
	cmp	_tb_use, 0		; if tb in shr empty
	je	cp_n_tx			; empty, need to cp and tx

; tb not empty, tx next frm in _tb_cur now
	call	_xmt_frm
	jmp	fill_tb

; tb empty, need to _cp_frm and then tx if there is any in tx_q 
cp_n_tx:
	mov	BX, OFFSET DGROUP:_tx_queue	; BX has offset of tx_q
	mov	AX, [BX]			; AX has 1st ptr
	cmp	AX, BX				; if point at self, q empty
	jne	get_tq
	jmp	rd_isr				; q empty. back to the top
		      				  
get_tq:
	mov	_tx_buf, AX			; q not empty
	mov	BP, SP
	push	AX				; PARM1: pointer to tx buf
	call	_QRemove				; call _QRemove
	mov	SP, BP
	 
	push	_tx_buf				; PARM1: pointer to tx buf
	call	_cp_frm				; call _cp_frm
	mov	SP, BP
	add	_tb_use, 1			

	call	_xmt_frm				; call _xmt_frm with no arg

; fill up the tb in shr mem as many as possible
fill_tb:
	mov	BX, OFFSET DGROUP:_tx_queue	; BX has offset of tx_q
	mov	AX, [BX]			; AX has 1st ptr
	cmp	AX, BX				; if point at self, q empty
	jne	tst_ful				; q not empty
	jmp	rd_isr				; q empty. back to the top
	
tst_ful:
	cmp	_tb_use,	tb_size			; if tb full
	jne	get_nxt				; not full
	jmp	rd_isr				; tb full. back to the top
		      				  
get_nxt:
	mov	_tx_buf, AX			; q not empty
	mov	BP, SP
	push	AX				; PARM1: pointer to tx buf
	call	_QRemove				; call _QRemove
	mov	SP, BP
	 
	push	_tx_buf				; PARM1: pointer to tx buf
	call	_cp_frm				; call _cp_frm
	mov	SP, BP
	add	_tb_use, 1			
	jmp	fill_tb				; loop for next one in q


;
; process counter overflow, update .algerr .crcerr .???(missed pkt)
;
test_cnt:
	test  	AL, MSK_CNT	
	jnz	prc_cnt			; yes, process cnt
	jmp	rd_isr	     		; no CNT irq, back to the top
; process CNT			
prc_cnt:
	mov	SI, OFFSET DGROUP:_MACstat	; SI is the base
	rd_wd8	CNTR0				; read CNTR0 in to AL
	xor	AH, AH
	add	WORD PTR [SI].algerr, AX
	adc	WORD PTR [SI].algerr + 2, 0	; update .algerr
	rd_wd8	CNTR1				; read CNTR1 in to AL
	xor	AH, AH
	add	WORD PTR [SI].crcerr, AX
	adc	WORD PTR [SI].crcerr + 2, 0	; update .crcerr
	rd_wd8	CNTR2				; read CNTR2 in to AL
	xor	AH, AH
;	add	WORD PTR [SI].???, AX
;	adc	WORD PTR [SI].??? + 2, 0	; update .???(missed)
	mov	AL, MSK_CNT
	wr_wd8	ISR			; clear CNT
	jmp	rd_isr			; back to the top

;
; no more irq
;
isr_exit:
; /* %%4%%+ */
	mov	al, EOI			;[6] 
	mov	dx, _int_reg		;[6] 
	out	dx, al			;[6]
	cmp	dx, INTB00		;[6] is it int controller 2
	jne	not_2
	mov	dx, INTA00		;[6] yes, it is controller 2
	out	dx, al			;[6] need to send EOI to controller 1 also
not_2:
					;
; /* %%4%%- */
	pop	rc			; save rc for FG				    
	mov	SS, CS:iintrss		; restore stack
	mov	SP, CS:iintrsp	
	
	cmp	_in_diags, 1
	je	skip_osi
	mov	_entersw,1
	call	_enter_osi

; restore registers
skip_osi:
	pop	SI
	pop	DI
	pop	BP
	pop	ES
	pop	DS
	pop	DX
	pop	CX
	pop	BX
	pop	AX

	iret

_hdw_isr	ENDP
 

	SUBTTL _cp_frm
	PAGE
; ******************************************************************
; CP_FRM(_tx_buf)
;
;
; Function:
;
; This routine copy a frame from system memory to shared memory tx
; buffer.
;
; tb_no = (_tb_cur + _tb_use) % TB_SIZE
; page_no = tb_no * TB_PGNO
;
;		TB_SIZE: number of _tx_buf's available in shared memory (2)					 
;		TB_PGNO: number of pages in a shared memory _tx_buf (3)
;
; C Call:	_cp_frm(_tx_buf)
;
; Calls:	_BuffFree
;
;
; On Entry:		
;	
;	PARM1  2 byte ptr to buffer header of the frame ready to be tx
; 	_WD8_base	base I/O address of board
;
; On Exit:
;	corrupt	AX,BX,CX,DX,DI  (ES should be reserved for bcopy routine!!!)
;
; Return Codes:								
;
;
; NOTES:
;
; ******************************************************************

	PUBLIC	_cp_frm 		; copy frame to shr mem

_cp_frm  PROC	NEAR					       

	push 	BP
	mov	BP, SP
	push	SI 
	push	DI   
	push	ES			; save ES
	
; get tb pointer
	mov	AL, _tb_use 
	add	AL, _tb_cur
	xor	AH, AH
	mov	BL, TB_SIZE		; AX/BL -> AL, AX%BL -> AH
	div	BL			; module in AH (tb number)
	mov	DL, AH			; save the tb number in DL
	mov	AL, AH
	mov	BL, TB_PGNO
	mul	BL			; AL*BL -> AX (page number in AX)		  
	mov	AH, AL
	xor	AL, AL			; tb offset in AX
	mov	DI, AX			; tb offset in dest DI
; set up dest seg in ES
	mov	AX, _SM_base
	mov	ES, AX			; shr seg in dest ES

; get parm1
	mov	BX, [BP].prm1		; _tx_buf in BX
	push 	BX			; save it for bufffree
	mov	BX, [BX].bd_addr	; ptr to start of data in BX
	mov	SI, BX			; blk lenth ptr by BX
	add	SI, 2			; data start in src SI(dest addr)
; copy dest addr from system _tx_buf
;	mov	CX, BPNA
;get_dst:	
;	mov	AL, DS:[SI]
;	mov	ES:[DI], AL
;	inc	DI
;	inc	SI
;	loop	get_dst
	mov	CX, (BPNA/2)
	rep	movsw			; copy destination addr 

; insert src addr to shr mem tb buff
;	mov	BP, _my_eaddr
;	mov	CX, BPNA
;get_src:
;	mov	AL, DS:[BP]
;	mov	ES:[DI], AL
;	inc	DI
;	inc	BP
;	loop	get_src		
	push	SI			; save system buff pointer for next movsw
	mov	SI, _my_eaddr
	mov	CX, (BPNA/2)
	rep	movsw			; insert source addr
	pop	SI			; restore SI for next movsw
	
; get txblk length						  
	mov	CX, [BX]		; txblk length in CX  
; write _tb_lnth for _xmt_frm
	mov	AX, CX
	add	AX, BPNA		; AX: byte count=txblk lngth + source addr
	mov	BX, OFFSET DGROUP:_tb_lnth
	xor	DH, DH
       	add	BX, DX			; DX has tb_no
	add	BX, DX			; BX has the offset if _tb_lnth(tb_no)
	mov	[BX], AX	 	; byte count(AX) into _tb_lnth(tb_no)

; copy the rest of the tx block
	sub	CX, BPNA		; length of txblk start after addr
	xor	AX, AX			; 0 in AX
	test	CX, 01h				 
	jz	even_b
	mov	AX, 1			; 1 in AX if odd no of byte
even_b:
	shr	CX, 1			; div by 2
	add	CX, AX			; word count in CX
	rep	movsw			; DS:SI -> ES:DI
; retrun _tx_buf
	pop	AX			; the _tx_buf
	mov	BP,SP
	push	AX			; parm1(_tx_buf)
	call	_BuffFree
	mov	SP,BP 

; exit now	
	pop	ES
	pop	DI
	pop	SI
	pop	BP	
	ret

_cp_frm  ENDP



	SUBTTL _xmt_frm
	PAGE
; **********************************************************************
; XMT_FRM()
;
; Function:
;
; This routine set up tx register and initial the tx process by a 8390
; command register. 
;
; C Call:	_xmt_frm ()
;
; Calls:
;
; On Entry:
;
; 	the frm resides in the _tb_cur tb of shared memory is the one to tx.
;
; On Exit:
;
; 	AX, BX corrupted
;
;
; Return Codes:
;
;
; ******************************************************************

	PUBLIC	_xmt_frm 		; Request frame transmit

_xmt_frm PROC	NEAR

	push	BP
	mov	BP, SP

; [05] do loop delay
	cmp	cs:starlan,0		;if (Starlan) {
	je	starlan_waitout
	push	cx
	mov	cx,cs:starlan_waitloopcnt ;. wait(starlan_waitloopcnt)
starlan_waitloop:
	loop	starlan_waitloop
	pop	cx
starlan_waitout:			;}
; end [05]


; get byte count from array _tb_lnth
	mov	AL, _tb_cur		; (tb number in AL)
	xor	AH, AH
	mov	BX, OFFSET DGROUP:_tb_lnth
       	add	BX, AX			; get the _tb_lnth(_tb_cur)
	add	BX, AX			;		
	mov	AX, [BX]	 	; byte count in AX

; set up TBCR0,1					 
	wr_wd8	TBCR0			; lower byte to TBCR0
	mov	AL, AH
	wr_wd8	TBCR1			; higher byte to TBCR1				

; get page number
	mov	AL, _tb_cur		; (tb number in AL)
	mov	BL, TB_PGNO
	mul	BL			; AL*BL -> AX (page number in AX)		  

; set up TPSR
	wr_wd8	TPSR			; write start page into TPSR

; issue tx command
	mov	AL, MSK_TXP + MSK_RD2
	wr_wd8	CMDR

; exit now
	mov	SP, BP
	pop	BP
	ret

_xmt_frm	ENDP




	SUBTTL _rcv_frm
	PAGE
; ******************************************************************
; RCV_FRM(rx_ptr)	
;
; Function:
;
; This routine copy a received frm in the shared mem into the system
; memory ,insert it into _rx_queue.
; up DMA controller for receive.
;
; C Call:	_rcv_frm(rx_ptr)
;
; Calls:	Buffalloc, _QInsert
;
; On Entry:
;
;   rx_ptr	Points to start address of rx frm in shard memory
; 
; On Exit:
;		destroy AX, BX, CX
;
; NOTES:
;
;
; ******************************************************************

	PUBLIC	_rcv_frm 		; enable receiver

_rcv_frm PROC	NEAR

	push	BP
	mov	BP, SP			
	push	SI
	push	DI
	push	ES
	
; get the rcv frm ptr from parameter
	mov	SI, [BP].prm1		; SI has ptr
	push	SI			; save ptr

	mov	SI, OFFSET DGROUP:_MACstat	; for stat update later
	push	SI				; save macstst offset

; call _BuffAlloc to get a system buffer
 	mov	BP, SP
	push	_bufsize	 		; parm 2
	mov	AX, OFFSET DGROUP:_bpool
	push	AX			; parm 1
	call	_BuffAlloc		; return value in AX
	mov	SP, BP
	
	or	AX, AX			; check return in AX
	jnz	has_buf			; continue   [02]
	jmp	no_buff			; AX = 0, no buffer

has_buf:                                ; [02] 
; got a buff, now update the .rxok
	pop	SI
	add	WORD PTR [SI].rxok, 1		; update .rxok
	adc	WORD PTR [SI].rxok + 2, 0			     

; copy from shr mem into the buff (AX)
; use rep movs to do the copy, DS:SI -> ES:DI	       
	mov	BX, AX			; BX has the prt to buffheader
	mov	DI, [BX].bd_addr	; dest offset in DI
	mov	AX, DS
	mov	ES, AX			; ES has dest seg ( system memory)
	mov	AX, _SM_base		
	mov	DS, AX			; src seg in DS	 (shr memory)
	pop	BP			; frm ptr in BP
	add	BP, 4
	mov	SI, BP			; src addr offset in SI
; read  byte count of data
	add	BP, BPNA		
	add	BP, BPNA		; BP ptr to data length filed	
	mov	CX, DS:[BP]		; CX has data byte count  
	mov	AL, CL
	mov	CL, CH
	mov	CH, AL			; exchange the cl and ch
	add	CX, RX_HDR_LEN		; add 12 byte for addrs
	add	CX, 2			; add 2 for data length field
; start [02]					 
; 8003E hw pass the CRC check with a invalid byte count??!!
; check if the byte count within the range <=_bufsize(=556=0x22c)
;	   

	mov	BP, OFFSET DGROUP:_bufsize
	cmp	CX, ES:[BP]
	jbe	lng_ok			; byte count <= 0x22c	[04]
	mov	AX, ES			; restore DS
	mov	DS, AX
	inc	_phreaque		; error! use _phreaque as error counter
	mov	BP, SP			; call _BuffFree to return OSI mem
	push	BX			; BX has the buff header
	call	_BuffFree		
	mov	SP, BP
	jmp	rcv_ret	 
		      
lng_ok:
; [02]
; get the word count form CX (byte count)
	xor	AX, AX			; 0 in AX
	test	CX, 01h			; CX has byte count				 
	jz	even_bt
	mov	AX, 1			; 1 in AX if odd byte
even_bt:
	shr	CX, 1			; div by 2
	add	CX, AX	 		; word count in CX
;
; check the wrap around case
;
	mov	AX, CX
	shl	AX, 1			; AX has byte count
	add	AX, SI			; AX has ending addr, if not wrap around
; /* %%3%%+ */
	cmp	AH, ES:stop_page	; [6]if out of boudary of shr mem
; /* %%3%%- */
	jge	wrap_a			; do wrap_around
	rep	movsw 			; no wrap_around, move string once			
	jmp	rstr_ds
;
; take care of the wrop around. may need to do move string twice
;
wrap_a:
; /* %%3%%+ */
	sub	AH, ES:stop_page	; [6]AX has byte cnt of 2nd movesw opr
; /* %%3%%- */ 
	shr	AX, 1			; AX has word cnt of 2nd movesw 
	sub	CX, AX			; CX has word cnt of 1st movesw opr
; 1st movesw
	rep	movsw			; DS:SI -> ES:DI
; 2nd movsw, if 2nd movsw needed
	cmp	AX, 0			
	je	rstr_ds			; no 2nd movsw
	mov	CX, AX			; word cnt in CX
	xor	AL, AL
	mov	AH, STRT_PG
	mov	SI, AX			; new starting addr in SI			
	rep	movsw			; DS:SI -> ES:DI 
; 
; restore DS
;
rstr_ds:
	mov	AX, ES
	mov	DS, AX

; put into rx_q
	mov	BP, SP
	mov	AX, OFFSET DGROUP:_rx_queue
	push	AX			; parm2	_rx_queue
	push	BX			; parm1 _rx_buf	
	call	_QInsert			
	mov	SP, BP
	jmp	rcv_ret
     							       
; no system buff availble to hold rx frm, update .rxnom 
no_buff:
	pop	SI			; put _MACstat offset in SI
	add	WORD PTR [SI].rxnom, 1		; update .rxnom
	adc	WORD PTR [SI].rxnom + 2, 0		      
	pop	SI 			; pop ptr						
					
rcv_ret:
 	pop	ES
	pop	DI
	pop	SI
	pop	BP
	ret

_rcv_frm ENDP
	

 

_TEXT	ENDS
	END

